#include "m_pd.h"
#include <octave/oct.h>
#include <octave/octave.h>
#include <octave/parse.h>
#include <octave/toplev.h>
#include <unistd.h>


static t_class *octave_class;
static int initted;

typedef struct _octave
{
    t_object x_obj;
    t_canvas *x_canvas;
} t_octave;

static void *octave_new(void)
{
    t_octave *x = (t_octave *)pd_new(octave_class);
    outlet_new(&x->x_obj, &s_);
    x->x_canvas = canvas_getcurrent();
    if (!initted)
    {
        string_vector argv (2);
        argv(0) = "embedded";
        argv(1) = "-q";
        octave_main (2, argv.c_str_vec (), 1);
        initted = 1;
    }
    return (x);
}

static void octave_eval(t_octave *x, t_symbol *s, int argc, t_atom *argv)
{
    char cmdbuf[1000];
    char atombuf[1000];
    int i, parse_status;
    cmdbuf[0] = 0;
    for (i = 0; i < argc; i++)
    {
    	strcat(cmdbuf, " ");
    	atom_string(&argv[i], atombuf, 1000);
    	strcat(cmdbuf, atombuf);
    }
    strcat(cmdbuf, ";\n");
    post("eval %s", cmdbuf);
    eval_string (cmdbuf, 0, parse_status);
    post("parse status %d", parse_status);
}

static void octave_source(t_octave *x, t_symbol *filesym)
{
    char buf[MAXPDSTRING], *bufptr, filename[MAXPDSTRING];
    int fd = open_via_path(canvas_getdir(x->x_canvas)->s_name,
         filesym->s_name, "", buf, &bufptr, MAXPDSTRING, 0);
    std::string warn_for;
    if (fd < 0)
    {
        pd_error(x, "%s: can't open file", filesym->s_name);
        return;
    }
    close(fd);
    strncpy(filename, buf, MAXPDSTRING);
    filename[MAXPDSTRING-2] = 0;
    strcat(filename, "/");
    strncat(filename, bufptr, MAXPDSTRING-strlen(filename));
    filename[MAXPDSTRING-1] = 0;
    post("file: %s", filename);
    source_file (filename, "", true, true, warn_for);
    post("warn_for: %s", warn_for.c_str());
}

static void octave_set(t_octave *x, t_symbol *s, int argc, t_atom *argv)
{
    if (argc < 2 || argv[0].a_type != A_SYMBOL)
    {
        pd_error(s, "set: usage: array-name value...");
        return;
    }
    Matrix mat(1, argc - 1);
    int i;
    for (i = 0; i < argc-1; i++)
        mat(0, i) = atom_getfloatarg(i+1, argc, argv);
    set_global_value(argv[0].a_w.w_symbol->s_name, mat);
}

static void octave_get(t_octave *x, t_symbol *var, t_symbol *selector)
{
    octave_value tmp = get_global_value (var->s_name, true);
    if (tmp.is_defined())
    {
        int i, totsize;
        t_atom *at;
        Matrix mat = tmp.matrix_value();
        dim_vector dims = mat.dims();
        int ndims = mat.ndims();
        post("ndims %d", ndims);
        for (i = 0, totsize = 1; i < ndims; i++)
            totsize *= dims.elem(i);
        post("size %d", totsize);
        mat.resize(1, totsize);
        at = (t_atom *)alloca(sizeof(t_atom) * totsize);
        for (i = 0; i < totsize; i++)
            SETFLOAT(at+i, mat(0, i));
        if (selector == &s_)
            outlet_list(x->x_obj.ob_outlet, 0, totsize, at);
        else outlet_anything(x->x_obj.ob_outlet, selector, totsize, at);
    }
    else post ("octave variable %s not found", var->s_name);
}

extern "C" {
void octave_setup(void)
{
    octave_class = class_new(gensym("octave"), (t_newmethod)octave_new, 0,
        sizeof(t_octave), 0, A_NULL);
    class_addmethod(octave_class, (t_method)octave_eval,
        gensym("eval"), A_GIMME, A_NULL);
    class_addmethod(octave_class, (t_method)octave_source,
        gensym("source"), A_SYMBOL, A_NULL);
    class_addmethod(octave_class, (t_method)octave_get,
        gensym("get"), A_SYMBOL, A_DEFSYM, A_NULL);
    class_addmethod(octave_class, (t_method)octave_set,
        gensym("set"), A_GIMME, A_NULL);
}
}
