/*

makenmf.c - NMF analsyis from 1 table into 2 others

*/

/*
NEXT: test this for convergence.  Then rearrange code (subroutines to
parse stuff and to do allloca-dependent stuff).  Then make a fake matrix
of rank (blah) for testing.
*/

#include "m_pd.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <lapacke/lapacke.h>

typedef struct _makenmf
{
    t_object x_ob;
    t_outlet *x_out1;
    t_outlet *x_out2;

    int x_debug;            /* debug flag */
} t_makenmf;

static t_class *makenmf_class;

    /* multiply matrix a by b, result into out.
    a is outrow x sharedim, b is sharedim x outcol,
        and out is outrow by outcol */
static void mpy_a_b(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[i * sharedim + k] * b[k * outcol + j];
        out[i * outcol + j] = accum;
    }
}

    /* multiply matrix (a transpose) by b, result into out.
    a is sharedim x outrow, b is sharedim x outcol,
        and out is outrow by outcol */
static void mpy_at_b(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[k * outrow + i] * b[k * outcol + j];
        out[i * outcol + j] = accum;
    }
}

    /* multiply matrix a by (b transpose), result into out.
    a is outrow x sharedim, b is outcol x sharedim,
        and out is outrow by outcol */
static void mpy_a_bt(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[i * sharedim + k] * b[j * sharedim + k];
        out[i * outcol + j] = accum;
    }
}


void test_matrix_fns( void)
{
    static double in22[4] = {1, 1, 0, 1}, in22b[4] = {1, 0, 1, 1},
    in12[2] = {1, 2}, out[100];
    mpy_a_b(out, in22, in22, 2, 2, 2);
    fprintf(stderr, "out 1:\n %f %f\n %f %f\n",
        out[0], out[1], out[2], out[3]);
    mpy_a_bt(out, in22, in22, 2, 2, 2);
    fprintf(stderr, "out 2:\n %f %f\n %f %f\n",
        out[0], out[1], out[2], out[3]);
    mpy_at_b(out, in22, in22, 2, 2, 2);
    fprintf(stderr, "out 3:\n %f %f\n %f %f\n",
        out[0], out[1], out[2], out[3]);
    mpy_at_b(out, in12, in22, 1, 2, 2);
    fprintf(stderr, "out 4:\n %f %f\n",
        out[0], out[1]);
    mpy_a_bt(out, in22, in12, 2, 1, 2);
    fprintf(stderr, "out 5:\n %f %f\n",
        out[0], out[1]);
}

static int preptabs(int argc, t_atom *argv,
    t_garray **ingarray, t_garray **wgarray, t_garray **hgarray, 
    t_word **invec, t_word **wvec, t_word **hvec, 
    int *nrows, int *ncolumns, int *nbasis)
{
    int insize, wsize, hsize;
    if (argc < 6 || argv[0].a_type != A_SYMBOL
        || argv[1].a_type != A_SYMBOL || argv[2].a_type != A_SYMBOL ||
            (*nrows = atom_getintarg(3, argc, argv)) <= 0 ||
            (*ncolumns = atom_getintarg(4, argc, argv)) <= 0 ||
            (*nbasis = atom_getintarg(5, argc, argv)) <= 0)
    {
        post("usage: makenmf_sqnorm intab htab wtab rows columns nbasis");
        return (0);
    }
    if (!(*ingarray = (t_garray *)pd_findbyclass(atom_getsymbolarg(0, argc, argv),
            garray_class)) ||
                !garray_getfloatwords(*ingarray, &insize, invec))
    {
        post("%s: not found", atom_getsymbolarg(0, argc, argv)->s_name);
        return (0);
    }
    if (!(*wgarray = (t_garray *)pd_findbyclass(atom_getsymbolarg(1, argc, argv),
            garray_class)) ||
                !garray_getfloatwords(*wgarray, &wsize, wvec))
    {
        post("%s: not found", atom_getsymbolarg(1, argc, argv)->s_name);
        return (0);
    }
    if (!(*hgarray = (t_garray *)pd_findbyclass(atom_getsymbolarg(2, argc, argv),
            garray_class)) ||
                !garray_getfloatwords(*hgarray, &hsize, hvec))
    {
        post("%s: not found", atom_getsymbolarg(2, argc, argv)->s_name);
        return (0);
    }
    if (insize < *nrows * *ncolumns)
    {
        post("table %s with %d elements too small to hold %dx%d matrix",
            atom_getsymbolarg(0, argc, argv)->s_name, insize, *nrows, *ncolumns);
        return (0);
    }
    if (hsize < *nbasis * *ncolumns)
    {
        post("table %s with %d elements too small to hold %dx%d matrix",
            atom_getsymbolarg(1, argc, argv)->s_name, hsize, *nbasis, *ncolumns);
        return (0);
    }
    if (wsize < *nrows * *nbasis)
    {
        post("table %s with %d elements too small to hold %dx%d matrix",
            atom_getsymbolarg(2, argc, argv)->s_name, wsize, *nrows, *nbasis);
        return (0);
    }
    return (1);
}

    /* iterate for square norm */
static void makenmf_sqnorm(t_makenmf *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k;
    double *vcopy;   /* input: nrows*ncolumns */
    double *wcopy;   /* nrows*nbasis */
    double *hcopy;   /* nbasis*ncolumns */
    double *wtvtmp;  /* w transpose * V: nbasis*ncolumns */
    double *whtmp;   /* w * h: nrows*ncolumns */
    double *wtwhtmp; /* wt * w * h: nbasis*ncolumns */
    double *vhttmp;   /* v * ht: nrows*nbasis */
    double *whhttmp;  /* w * h * ht: nrows*nbasis */
    int totaltmpspace;
    double errtotal, meanchange;

    t_garray *ingarray, *wgarray, *hgarray;
    t_word *invec, *wvec, *hvec;
    int nrows, ncolumns, nbasis;
    if (!preptabs(argc, argv, &ingarray, &wgarray, &hgarray, 
        &invec, &wvec, &hvec, &nrows, &ncolumns, &nbasis))
            return;

    totaltmpspace =
        2 * nrows*ncolumns + 3 * nrows*nbasis + 3*nbasis*ncolumns;
    vcopy = (double *)alloca(sizeof(double) * totaltmpspace);
    wcopy = vcopy + nrows*ncolumns;
    hcopy = wcopy + nrows*nbasis;
    wtvtmp = hcopy + nbasis*ncolumns;
    whtmp = wtvtmp + nbasis*ncolumns;
    wtwhtmp = whtmp + nrows*ncolumns;
    vhttmp = wtwhtmp + nbasis*ncolumns;
    whhttmp = vhttmp + nrows*nbasis;
    
    for (i = 0; i < nrows*ncolumns; i++)
        vcopy[i] = invec[i].w_float;
    for (i = 0; i < nrows*nbasis; i++)
        wcopy[i] = wvec[i].w_float;
    for (i = 0; i < nbasis*ncolumns; i++)
        hcopy[i] = hvec[i].w_float;
    mpy_a_b(whtmp, wcopy, hcopy, nrows, ncolumns, nbasis);
    mpy_at_b(wtvtmp, wcopy, vcopy, nbasis, ncolumns, nrows);
    mpy_at_b(wtwhtmp, wcopy, whtmp, nbasis, ncolumns, nrows);
    mpy_a_bt(vhttmp, vcopy, hcopy, nrows, nbasis, ncolumns);
    mpy_a_bt(whhttmp, whtmp, hcopy, nrows, nbasis, ncolumns);
    
    errtotal = 0;
    for (i = 0; i < nrows * ncolumns; i++)
    {
        double diff =  whtmp[i] - vcopy[i];
        errtotal += diff * diff;
    }
    post("RMS error %f", sqrt(errtotal/(nrows * ncolumns)));

    errtotal = 0;
    for (i = 0; i < nrows * nbasis; i++)
    {
        double new, old = wcopy[i], diff;
        if (whhttmp[i] <= 0)
        {
            post("zero denominator skipped");
            new = old;
        }   
        else new = old * (0.25 + 0.75 * vhttmp[i] / whhttmp[i]);
        diff = new - old;
        errtotal += diff * diff;
        wvec[i].w_float = new;
    }
    for (i = 0; i < nbasis * ncolumns; i++)
    {
        double new, old = hcopy[i], diff;
        if (wtwhtmp[i] <= 0)
        {
            post("zero denominator skipped");
            new = old;
        }   
        else new = old * (0.25 + 0.75 * wtvtmp[i] / wtwhtmp[i]);
        diff = new - old;
        errtotal += diff * diff;
        hvec[i].w_float = new;
    }
    errtotal /= nbasis * (nrows * ncolumns);
    meanchange = sqrt(errtotal);
    post("mean change is %f", meanchange);
    garray_redraw(hgarray);
    garray_redraw(wgarray);
}

    /* initialize with random H and W */
static void makenmf_init(t_makenmf *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k;
    double average, average2, range;

    t_garray *ingarray, *wgarray, *hgarray;
    t_word *invec, *wvec, *hvec;
    int nrows, ncolumns, nbasis;
    if (!preptabs(argc, argv, &ingarray, &wgarray, &hgarray, 
        &invec, &wvec, &hvec, &nrows, &ncolumns, &nbasis))
            return;

    for (i = 0, average = 0; i < nrows * ncolumns; i++)
        average += invec[i].w_float;
    average /= nrows * ncolumns ;
    range = 2 * sqrt(average / nbasis);
    for (i = 0; i < nbasis * ncolumns; i++)
        hvec[i].w_float = range * (1./RAND_MAX) * random();
    for (i = 0; i < nrows * nbasis; i++)
        wvec[i].w_float = range * (1./RAND_MAX) * random();
    for (i = 0, average2 = 0; i < nrows; i++)   
        for (j = 0; j < ncolumns; j++)   
            for (k = 0; k < nbasis; k++)
    {
        average2 += wvec[i * nbasis + k].w_float *
            hvec[k * ncolumns + j].w_float;
    }
    average2 /=  nrows * ncolumns;
    post("average %f, average2 = %f", average, average2);
    garray_redraw(hgarray);
    garray_redraw(wgarray);
}


    /* put a fake input in that will factor perfectly */
static void makenmf_fakedata(t_makenmf *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k;

    t_garray *ingarray, *wgarray, *hgarray;
    t_word *invec, *wvec, *hvec;
    int nrows, ncolumns, nbasis;
    if (!preptabs(argc, argv, &ingarray, &wgarray, &hgarray, 
        &invec, &wvec, &hvec, &nrows, &ncolumns, &nbasis))
            return;

    for (i = 0; i < nrows * nbasis; i++)
        wvec[i].w_float = (1./RAND_MAX) * random();
    for (i = 0; i < nbasis * ncolumns; i++)
        hvec[i].w_float = (1./RAND_MAX) * random();
    for (i = 0; i < nrows; i++)
        for (j = 0; j < ncolumns; j++)
    {
        double sum = 0;
        for (k = 0; k < nbasis; k++)
            sum += wvec[nbasis*i + k].w_float * hvec[ncolumns*k + j].w_float;
        invec[i * ncolumns + j].w_float = sum;
    }
    garray_redraw(hgarray);
    garray_redraw(wgarray);
    garray_redraw(ingarray);
}


    /* factor matrix H into optimal W (decompose 'nrows' rows of V as linear
    combinations of the 'nbasis' rows of H , puting result in W.  Each row
    of W is one decomposition. */
static void makenmf_factor(t_makenmf *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k, laout;
    double *vcopy;   /* input: nrows*ncolumns */
    double *wcopy;   /* nrows*nbasis */
    double *hcopy;   /* nbasis*ncolumns */
    double *hhttmp;  /* H * H transpose: nbasis*nbasis */
    double *vhttmp;   /* V * H transpose: nrows*nbasis */
    double *answertmp; /* output of linpack routine: nrows*nbasis */
    lapack_int *lpiv;   /* stuff for PAPACK, nbasis ints */
    double errtotal;
    
    t_garray *ingarray, *wgarray, *hgarray;
    t_word *invec, *wvec, *hvec;
    int nrows, ncolumns, nbasis;
    if (!preptabs(argc, argv, &ingarray, &wgarray, &hgarray, 
        &invec, &wvec, &hvec, &nrows, &ncolumns, &nbasis))
            return;

    vcopy = (double *)alloca(
        sizeof(double) * (nrows*ncolumns + nrows*nbasis + nbasis*ncolumns +
            nbasis*nbasis + nrows*nbasis + nrows*nbasis) +
        sizeof(lapack_int) * nbasis);
    wcopy = vcopy + nrows*ncolumns;
    hcopy = wcopy + nrows*nbasis;
    hhttmp = hcopy + nbasis*ncolumns;
    vhttmp = hhttmp + nbasis*nbasis;
    answertmp = vhttmp + nrows*nbasis;
    lpiv = (lapack_int *)(answertmp + nrows*nbasis);

    for (i = 0; i < nrows*ncolumns; i++)
        vcopy[i] = invec[i].w_float;
    for (i = 0; i < nbasis*ncolumns; i++)
        hcopy[i] = hvec[i].w_float;

    mpy_a_bt(hhttmp, hcopy, hcopy, nbasis, nbasis, ncolumns);
    mpy_a_bt(vhttmp, vcopy, hcopy, nrows, nbasis, ncolumns);
    laout = LAPACKE_dsysv(LAPACK_COL_MAJOR, 'u',
        nbasis, nrows, hhttmp, nbasis, lpiv, vhttmp, nbasis);
    if (laout != 0)
    {
        post("lapack returned error %d", laout);
        return;
    }
    for (i = 0; i < nrows * nbasis; i++)
        wvec[i].w_float = vhttmp[i];

    garray_redraw(wgarray);
}

static void makenmf_debug(t_makenmf *x, t_float f)
{
    x->x_debug = f;
}

static void makenmf_print(t_makenmf *x, t_float f)
{
    int i, j;
    post("debug %d", x->x_debug);
}

static void makenmf_free(t_makenmf *x)
{
}

static void *makenmf_new(t_symbol *s)
{
    int i, j;
    t_makenmf *x = (t_makenmf *)pd_new(makenmf_class);
    x->x_out1 = outlet_new(&x->x_ob, &s_list);
    x->x_out2 = outlet_new(&x->x_ob, &s_list);
    x->x_debug = 0;
    return (x);
}

void makenmf_setup(void)
{
    makenmf_class = class_new(gensym("makenmf"),
        (t_newmethod)makenmf_new, (t_method)makenmf_free,
	sizeof(t_makenmf), 0, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_init,
        gensym("init"), A_GIMME, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_fakedata,
        gensym("fakedata"), A_GIMME, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_sqnorm,
        gensym("sqnorm"), A_GIMME, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_factor,
        gensym("factor"), A_GIMME, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_print,
        gensym("print"), A_DEFFLOAT, 0);
    class_addmethod(makenmf_class, (t_method)makenmf_debug,
        gensym("debug"), A_FLOAT, 0);
}

