/* searchvec -- experimental object to search for a vector. */

/*  Copyright 2004 Miller Puckette.
Permission is granted to use this software for any purpose provided you
keep this copyright notice intact.

THE AUTHOR AND HIS EMPLOYERS MAKE NO WARRANTY, EXPRESS OR IMPLIED,
IN CONNECTION WITH THIS SOFTWARE.
 
This file may someday be downloadable from http://www.crca.ucsd.edu/~msp .
*/

char version[] = "searchvec version 0.2";

#include "m_pd.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
static t_class *searchvec_class;
#define VECDIM 3
#define BIGVECDIM (VECDIM+1)
#define VECBYTES (sizeof(float)*BIGVECDIM)
#define FILEVECDIM 11
#define CONTROLDIM 5
#define NEIGHBOR 2      /* number of steps in any direction considered local */

typedef struct _searchvec
{
    t_object x_obj;
    t_float *x_vec;
    int x_nvec;
    float x_ave[BIGVECDIM];
    float x_norm[VECDIM][VECDIM];
    int x_dims[CONTROLDIM];
    int x_ncontroldim;
    int x_nrandom;
    float x_lastout[CONTROLDIM];
    t_outlet *x_out1;
    t_outlet *x_out2;
} t_searchvec;

    /* normalized cosine transform.  Input has FILEVECDIM points; output
    only the first BIGVECDIM. */
static void cosinetransform(float *out, float *in)
{
    int j, k;
    float sum, correct = 1./sqrt((float)FILEVECDIM),
        angle = 3.14159/(float)FILEVECDIM;
    for (k = 0; k < BIGVECDIM; k++)
    {
        for (sum = 0, j = 0; j < FILEVECDIM; j++)
	    sum += cos(angle*j*k) * in[j];
        out[k] = sum * correct;
    }
}

static void *searchvec_new(t_float fnonrepeat)
{
    t_searchvec *x = (t_searchvec *)pd_new(searchvec_class);
    int i;
    x->x_out1 = outlet_new(&x->x_obj, gensym("list"));
    x->x_out2 = outlet_new(&x->x_obj, gensym("float"));
    x->x_vec = (t_float *)getbytes(0);
    x->x_nvec = 0;
    x->x_nrandom = 0;
    for (i = 0; i < BIGVECDIM; i++)
    {
    	x->x_ave[i] = 0;
    }
    for (i = 0; i < CONTROLDIM; i++)
    {
    	x->x_dims[i] = 1;
	x->x_lastout[i] = -1;
    }
    x->x_dims[0] = 0;
    x->x_ncontroldim = 1;
    return (x);
}

static void searchvec_read(t_searchvec *x, t_symbol *s)
{
    FILE *fd;
    float invec[FILEVECDIM], z;
    int i, nvec = 0, remaining;
    float *fp, *fp2;

    int ifd;
    char namebuf[MAXPDSTRING], *nb2;

    if ((ifd = open_via_path(".", s->s_name, "", 
    	namebuf, &nb2, MAXPDSTRING, 0)) < 0)
    {
	post("%s: couldn't open", s->s_name);
	return;
    }
    fd = fdopen(ifd, "r");

    if (!fd)
    {
	post("%s: open failed", s->s_name);
	return;
    }
    x->x_vec = (float *)t_resizebytes(x->x_vec, 
	x->x_nvec * VECBYTES, 0);
    while (1)
    {
	for (i = FILEVECDIM, fp = invec; i--; fp++)
	    if (fscanf(fd, "%f", fp) < 1) goto nomore;
	x->x_vec = (float *)t_resizebytes(x->x_vec,
	    nvec * VECBYTES,
	    (nvec + 1) * VECBYTES);
        cosinetransform(&x->x_vec[nvec*BIGVECDIM], invec);
	nvec++;
    }
nomore:
    if (i != FILEVECDIM-1)
    	post("searchvec: %d extra numbers?", FILEVECDIM-1-i);
    post("searchvec: read %d vectors", nvec);
    x->x_nvec = nvec;
    fclose(fd);
#if 0
    post("raw:");
    searchvec_print(x);
#endif
#if 0
    for (i = 0, z = 0; i < BIGVECDIM; i++)
    	z += x->x_vec[0+i] * x->x_vec[0+i];
    post("mag %f", z);
    for (i = 0, z = 0; i < BIGVECDIM; i++)
    	z += x->x_vec[BIGVECDIM+i] * x->x_vec[BIGVECDIM+i];
    post("mag %f", z);
    for (i = 0, z = 0; i < BIGVECDIM; i++)
    	z += x->x_vec[0+i] * x->x_vec[BIGVECDIM+i];
    post("xprod %f", z);
    for (i = 0, z = 0; i < BIGVECDIM; i++)
    	z += x->x_vec[0+i] * x->x_vec[2*BIGVECDIM+i];
    post("xprod %f", z);
#endif
    	/* assume 1-dimensional for now.  "geometry" can re-specify this. */
    for (i = 0; i < CONTROLDIM; i++)
    	    x->x_dims[i] = 1;
    x->x_dims[0] = x->x_nvec;
    x->x_ncontroldim = 1;

}

static void searchvec_writestats(t_searchvec *x, t_symbol *s1, t_symbol *s2)
{
    FILE *fd;
    double cor[VECDIM][VECDIM];
    double ave[BIGVECDIM];
    int i, j,k;

    for (i = 0; i < BIGVECDIM; i++)
    {
    	double sum = 0;
    	for (j = 0; j < x->x_nvec; j++)
    	    sum += x->x_vec[j * BIGVECDIM + i];
	ave[i] = sum / x->x_nvec;
    }
    for (i = 0; i < VECDIM; i++)
    	for (j = 0; j < VECDIM; j++)
    {
    	double sum = 0;
    	for (k = 0; k < x->x_nvec; k++)
    	    sum += (x->x_vec[k * BIGVECDIM + i + 1] - ave[i+1]) *
	    	(x->x_vec[k * BIGVECDIM + j + 1] - ave[j+1]);
	cor[i][j] = sum / x->x_nvec;
    }
    
    if (!(fd = fopen(s1->s_name, "w")))
    {
	post("%s: couldn't open for writing", s1->s_name);
	return;
    }
    for (i = 0; i < BIGVECDIM; i++)
    	fprintf(fd, "%f\n", (float)ave[i]);
    fclose(fd);

    if (!(fd = fopen(s2->s_name, "w")))
    {
	post("%s: couldn't open for writing", s2->s_name);
	return;
    }
    for (i = 0; i < VECDIM; i++)
    {
    	for (j = 0; j < VECDIM; j++)
    	    fprintf(fd, "%f ", (float)cor[i][j]);
    	fprintf(fd, "\n");
    }    
    fclose(fd);

}

static void searchvec_normalize(t_searchvec *x, t_symbol *s1, t_symbol *s2)
{
    FILE *fd;
    double norm[VECDIM][VECDIM];
    double ave[BIGVECDIM];
    int i, j, k, bad = 0;
    int ifd;
    char namebuf[MAXPDSTRING], *nb2;


    if ((ifd = open_via_path(".", s1->s_name, "", 
    	namebuf, &nb2, MAXPDSTRING, 0)) < 0)
    {
	post("%s: couldn't open", s1->s_name);
	return;
    }
    fd = fdopen(ifd, "r");

    for (i = 0; i < BIGVECDIM; i++)
    {
    	if (fscanf(fd, "%lf\n", &ave[i]) < 1)
	{
	     post("%s: read failed", s1->s_name);
	     bad = 1;
	     break;
    	}
    }
    fclose(fd);

    if ((ifd = open_via_path(".", s2->s_name, "", 
    	namebuf, &nb2, MAXPDSTRING, 0)) < 0)
    {
	post("%s: couldn't open", s2->s_name);
	return;
    }
    fd = fdopen(ifd, "r");

    for (i = 0; i < VECDIM; i++)
    	for (j = 0; j < VECDIM; j++)
    {
    	if (fscanf(fd, "%lf\n", &norm[i][j]) < 1)
	{
	     post("%s: read failed", s1->s_name);
	     bad = 1;
	     break;
    	}
    }
    fclose(fd);
    if (bad)
    	return;

    for (i = 0; i < x->x_nvec; i++)
    	for (j = 0; j < BIGVECDIM; j++)
	    x->x_vec[i*BIGVECDIM + j] -= ave[j];

    for (i = 0; i < x->x_nvec; i++)
    {
    	double outvec[VECDIM];
    	for (j = 0; j < VECDIM; j++)
	{
	    outvec[j] = 0;
	    for (k = 0; k < VECDIM; k++)
	    	outvec[j] += norm[j][k]*x->x_vec[i*BIGVECDIM+k+1];
    	}
    	for (j = 0; j < VECDIM; j++)
	    x->x_vec[i*BIGVECDIM+j+1] = outvec[j];
    }
}

static void searchvec_readstats(t_searchvec *x, t_symbol *s1, t_symbol *s2)
{
    FILE *fd;
    int i, j, k, bad = 0;
    int ifd;
    char namebuf[MAXPDSTRING], *nb2;

    if ((ifd = open_via_path(".", s1->s_name, "", 
    	namebuf, &nb2, MAXPDSTRING, 0)) < 0)
    {
	post("%s: couldn't open", s1->s_name);
	return;
    }
    fd = fdopen(ifd, "r");
    for (i = 0; i < BIGVECDIM; i++)
    {
    	if (fscanf(fd, "%f\n", &x->x_ave[i]) < 1)
	{
	     post("%s: read failed", s1->s_name);
	     bad = 1;
	     break;
    	}
    }
    fclose(fd);

    if ((ifd = open_via_path(".", s2->s_name, "",
    	namebuf, &nb2, MAXPDSTRING, 0)) < 0)
    {
	post("%s: couldn't open", s2->s_name);
	return;
    }
    fd = fdopen(ifd, "r");
    for (i = 0; i < VECDIM; i++)
    	for (j = 0; j < VECDIM; j++)
    {
    	if (fscanf(fd, "%f\n", &x->x_norm[i][j]) < 1)
	{
	     post("%s: read failed", s1->s_name);
	     bad = 1;
	     break;
    	}
    }
    fclose(fd);
}

static void searchvec_clear(t_searchvec *x)
{
    x->x_vec = (float *)resizebytes(x->x_vec, x->x_nvec * VECBYTES, 0);
    x->x_nvec = 0;
}

static void searchvec_print(t_searchvec *x, t_float f)
{
    post("nrandom %d", x->x_nrandom);
    if (f > 0)
    {
        int j;
        for (j = 0; j < x->x_nvec; j++)
        {
    	    float *w = x->x_vec + j * BIGVECDIM + 1;
    	    post("a %5.2f; \
w %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f",
    	    x->x_vec[j * BIGVECDIM], w[0], w[1], w[2], w[3], w[4], w[5],
    	    	w[6], w[7], w[8], w[9]);
        }
    }
}

static void searchvec_params(t_searchvec *x,
    t_symbol *s, int argc, t_atom *argv)
{
    x->x_nrandom = atom_getfloatarg(0, argc, argv);
}

static void searchvec_geometry(t_searchvec *x,
    t_symbol *s, int argc, t_atom *argv)
{
    int i, prod;
    if (argc > CONTROLDIM)
    {
    	post("searchvec: too many dimensions");
	goto bad;
    }
    for (i = 0, prod = 1; i < argc; i++)
    {
    	x->x_dims[i] = atom_getfloatarg(i, argc, argv);
	if (x->x_dims[i] < 3)
	{
    	    post("searchvec: need at least 3 in every dimension");
	    goto bad;
	}
	prod *= x->x_dims[i];
    }
    if (prod != x->x_nvec)
    {
    	post("searchvec:  dimension product doesn't equal nvec");
	goto bad;
    }
    x->x_ncontroldim = argc;
    return;
bad:
    for (i = 0; i < CONTROLDIM; i++)
    	    x->x_dims[i] = 1;
    x->x_dims[0] = x->x_nvec;
    x->x_ncontroldim = 1;
    
}

static void searchvec_forget(t_searchvec *x)
{
    int i;
    for (i = 1; i < x->x_ncontroldim; i++)
    	x->x_lastout[i] = -1;
}

static float searchvec_dist(float *vec, float *normvec)
{
    int j;
    float dist;
    for (dist = 0, j = 0; j < VECDIM; j++)
	dist += (vec[1 + j] - normvec[j]) *
	    (vec[1 + j] - normvec[j]);
    return (dist);
}

static void searchvec_list(t_searchvec *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k, prod;
    float closest = 1e20;
    float invec[FILEVECDIM];
    float rotvec[BIGVECDIM], normvec[BIGVECDIM];
    float amp;
    t_atom controlout[CONTROLDIM];
    float frac;
    int countvec[CONTROLDIM], skipindex[CONTROLDIM],
        searchstart[CONTROLDIM], searchn[CONTROLDIM];
    int bestindex = 0, digitindex, interpindex, bestfloatindex,
    	interpfloatindex;

    for (i = 0; i < FILEVECDIM; i++)
    	invec[i] = atom_getfloatarg(i, argc, argv);
#if 0
    for (i = 0; i < BIGVECDIM; i++)
    	post("%f", invec[i]);
    post("");
#endif
    if (x->x_nvec < 3)
    {
    	post("searchvec: need at least 3 vectors");
	return;
    }
    cosinetransform(rotvec, invec);

    for (j = 0; j < BIGVECDIM; j++)
	rotvec[j] -= x->x_ave[j];
#if 0
    for (i = 0; i < BIGVECDIM; i++)
    	post("%f", rotvec[i]);
    post("");
#endif

    for (j = 0; j < VECDIM; j++)
    {
	normvec[j] = 0;
	for (k = 0; k < VECDIM; k++)
	    normvec[j] += x->x_norm[j][k]*rotvec[k+1];
#if 0
    	post("%f", normvec[j]);
#endif
    }
    if (x->x_nrandom <= 0 || x->x_lastout[0] < 0)
        /* OLD WAY, exhaustively searching through everything */
    {
        for (i = 0, prod = BIGVECDIM; i < x->x_ncontroldim; i++)
        {
    	    countvec[i] = 0;
	    skipindex[i] = prod;
    	    prod *= x->x_dims[i];
	    if (x->x_dims[i] < 3)
	    {
    	        post("searchvec: need at least 3 in every dimension");
	        return;
	    }
        }

    	    /* search for best match.  LATER, we'll add hysteresis here, which
	    will mean keeping track of all indices as we go. */
        for (i = 0; i < x->x_nvec; i++)
        {
    	    float dist = searchvec_dist(&x->x_vec[i * BIGVECDIM], normvec);
    	        /* see how far we are from previous outout */
#if 0 /* hysteresis, commented out */
	    if (x->x_maxdistance > 0 && x->x_lastout[0] >= 0)
	    {
	        float extradist = 0;
	        for (j = 0; j < x->x_ncontroldim; j++)
	        {
		    float movement = x->x_lastout[j] - countvec[j];
		    if (movement < 0)
		        movement = -movement;
	    	    if (movement > 0.5)
		        movement -= 0.5;
		    else movement = 0;
		    extradist += x->x_metric[j] * movement;
	        }
	        if (extradist > x->x_maxdistance)
	    	    extradist = x->x_maxdistance;
	        dist += extradist;
	    }
#endif /* hysteresis, commented out */

	    if (dist < closest)
	        closest = dist, bestindex = i;

	    for (j = 0; j < x->x_ncontroldim; j++)
	    {
	        if (++countvec[j] == x->x_dims[j])
	    	    countvec[j] = 0;
	        else break;
	    }
        }
    }
    else    /* nrandom is positive and the last point is valid; search neighbors plus nrandom new spots */
    {
        int ntodo = 1;
            /* if there's a previous output, search its neighbors */
        for (i = 0, prod = 1; i < x->x_ncontroldim; i++)
        {
    	    countvec[i] = 0;
	    skipindex[i] = prod;
            if (x->x_dims[i] < 2*NEIGHBOR+1)
            {
                searchn[i] = x->x_dims[i];
                searchstart[i] = 0;
            }
            else
            {
                searchn[i] = 2*NEIGHBOR+1;
                if (x->x_lastout[i] < NEIGHBOR)
                    searchstart[i] = 0;
                else if (x->x_lastout[i] - NEIGHBOR >=
                    x->x_dims[i] - (2*NEIGHBOR+1))
                        searchstart[i] = x->x_dims[i] - (2*NEIGHBOR+1);
                else searchstart[i] = x->x_lastout[i] - NEIGHBOR;
            }
            ntodo *= searchn[i];
    	    prod *= x->x_dims[i];   /* LATER look at this... */
        }
        /* post("start %d", dim %d,  prod); */
        for (i = 0; i < ntodo; i++)
        {
            int tryindex;
            float newresult;
            float distance;
            for (j = tryindex = 0; j < x->x_ncontroldim; j++)
                tryindex += skipindex[j] * (searchstart[j] + countvec[j]);
            newresult =
                searchvec_dist(&x->x_vec[tryindex * BIGVECDIM], normvec);
            if (newresult < closest)
            {
                closest = newresult;
                bestindex = tryindex;
            }
	    for (j = 0; j < x->x_ncontroldim; j++)
	    {
	        if (++countvec[j] == x->x_dims[j])
	    	    countvec[j] = 0;
	        else break; 
            }
        }

    	    /* search randomly for more ideas */
        for (i = 0; i < x->x_nrandom; i++)
        {
            int tryindex = random() * (1./(float)RAND_MAX) * x->x_nvec;
            float dist;
            if (tryindex >= x->x_nvec)
                tryindex = 0;

    	    dist = searchvec_dist(&x->x_vec[tryindex * BIGVECDIM], normvec);

	    if (dist < closest)
	        closest = dist, bestindex = tryindex;
        }
    }       /* end of if (x->x_nrandom...) */

    bestfloatindex = bestindex * BIGVECDIM;
    digitindex = bestindex;

#if 0
    post("best distance %f",  closest);
#endif

    	/* now "interpolate" across each dimension to get fractionally
	accurate control values */
    for (i = 0; i < x->x_ncontroldim; i++)
    {
    	int subindex = digitindex % x->x_dims[i];
	float f1, f2, middle;
	digitindex = digitindex / x->x_dims[i];
	if (subindex < 1)
    	    interpindex = 1, interpfloatindex = bestfloatindex + skipindex[i];
	else if (subindex == x->x_dims[i]-1)
    	    interpindex = x->x_dims[i]-2, 
	    	interpfloatindex = bestfloatindex - skipindex[i];
	else interpindex = subindex, interpfloatindex = bestfloatindex;

    	f1 = searchvec_dist(&x->x_vec[interpfloatindex - skipindex[i]],
	    normvec);
    	f2 = searchvec_dist(&x->x_vec[interpfloatindex + skipindex[i]],
	    normvec);
    	middle = searchvec_dist(&x->x_vec[interpfloatindex], normvec);
	    /* dunno why we square the errors here; this works best
	    	empirically */
	f1 = f1*f1;
	f2 = f2*f2;
	middle = middle*middle;
	float den = f1+f2 - 2 * middle;
	if (den < 1e-20)
	    frac = 0;
	    	/* vertex of parabola passing through the three points */
	else frac = (f1-f2)/(2*den);
#if 0
	post("dists %f %f %f", f1, middle, f2);
	post("interp %d, raw frac %f", interpindex, frac);
#endif
	if (subindex < 1)
	{
	    if (den < 0)
	    	frac = -1;
	    else if (frac < -1)
	    	frac = -1;
	    else if (frac > 0.5)
	    	frac = 0.5;
	}
	else if (subindex == x->x_dims[i]-1)
	{
	    if (den < 0)
	    	frac = 1;
	    else if (frac < -0.5)
	    	frac = -0.5;
	    else if (frac > 1)
	    	frac = 1;
	}
	else if (frac < -0.5)
    	    frac = -0.5;
	else if (frac > 0.5)
    	    frac = 0.5;
	SETFLOAT(&controlout[i], x->x_lastout[i] = interpindex + frac);
    }

    outlet_float(x->x_out2, rotvec[0] - x->x_vec[bestindex * BIGVECDIM]);
    outlet_list(x->x_out1, 0, x->x_ncontroldim, controlout);
}

static void searchvec_free(t_searchvec *x)
{
    freebytes(x->x_vec, x->x_nvec * VECBYTES);
}

void searchvec_setup(void)
{
    searchvec_class = class_new(gensym("searchvec"),
    	(t_newmethod)searchvec_new, (t_method)searchvec_free,
	sizeof(t_searchvec), 0, A_DEFFLOAT, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_read,
    	gensym("read"), A_SYMBOL, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_writestats,
    	gensym("writestats"), A_SYMBOL, A_SYMBOL, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_readstats,
    	gensym("readstats"), A_SYMBOL, A_SYMBOL, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_normalize,
    	gensym("normalize"), A_SYMBOL, A_SYMBOL, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_geometry,
    	gensym("geometry"), A_GIMME, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_params,
    	gensym("params"), A_GIMME, 0);
    class_addmethod(searchvec_class, (t_method)searchvec_forget,
    	gensym("forget"), 0);
    class_addmethod(searchvec_class, (t_method)searchvec_print,
    	gensym("print"), A_DEFFLOAT, 0);
    class_addlist(searchvec_class, searchvec_list);
    post(version);
}
